package ActiveState::ProcessPool;

use strict;
use Data::Dump 'dump';

use base qw(DynaLoader Exporter);

our $VERSION = '1.00';
our @EXPORT_OK = qw(encode_kv decode_kv);

ActiveState::ProcessPool->bootstrap($VERSION);

sub restart {
    my ($self, $slot, $func) = @_;
    $self->stop($slot);
    $self->start($slot, $func);
}

sub send {
    my ($self, $slot, @kv) = @_;
    my $fd = $self->wfd($slot);
    $self->inc($slot);
    sendfd($fd, @kv);
}

sub recv {
    my ($self, $slot) = @_;
    my $fd = $self->rfd($slot);
    recvfd($fd);
}

sub start {
    my ($self, $slot, $callback) = @_;
    $self->start_long($slot, sub { _long($callback) });
}

sub _long {
    my $callback = shift;
    select(STDERR); # unqualified print()s go to STDERR
    while (1) {
	eval {
	    my @kv = recvfd(0);
	    my @kv2 = &$callback(@kv);
	    sendfd(1, @kv2);
	};
	if ($@) {
	    warn $@;
	    last; # child will exit
	}
    }
}

sub sendfd {
    my ($fd, @kv) = @_;
    my $data = encode_kv(@kv);
    my $pack = pack('N', length $data) . $data;
    writefd($fd, $pack);
}

sub recvfd {
    my $fd = shift;
    my $pack;
    my $read = readfd($fd, $pack, 2048, 0);
    return if $read < 4;
    # length size + data size
    my $size = unpack('N', $pack) + 4;
    while ($read < $size) {
	my $more;
	my $want = $size - $read;
	$more = readfd($fd, $pack, $want, $read)
	    or die "can't read $want bytes: got $more: $!";
	$read += $more;
    }
    decode_kv(substr($pack, 4));
}

sub write {
    my ($self, $slot, $msg) = @_;
    my $fd = $self->wfd($slot);
    writefd($fd, $msg);
}

sub read {
    my $self = shift;
    my $slot = shift;
    my $fd = $self->rfd($slot);
    readfd($fd, @_);
}

1;

=head1 NAME

ActiveState::ProcessPool - run a subroutine in a pool of processes

=head1 SYNPOSIS

   # Allocate a pool of up to 10 processes
   $pool = ActiveState::ProcessPool->new(10);

   # Start a slot that will invoke a subroutine each time the
   # parent sends it a request. ActiveState::ProcessPool will
   # marshall data on the input and output for you.
   for $slot (0 .. $pool->size) {
      $pool->start($slot, \&runner);
   }

   # Start a slot that will invoke a long-running subroutine
   # that can parse ActiveState::ProcessPool's input/output
   # requirements.
   for $slot (0 .. $pool->size) {
      $pool->start_long($slot, \&longrun);
   }

   # Invoke process zero, passing it the key/value pairs
   # specified:
   $pool->write($slot, a => 1, b => 2, c => 3);

   # ... wait a while

   my @response = $pool->read($slot);

=head1 DESCRIPTION

ActiveState::ProcessPool allows you work in parallel by invoking a subroutine
using a pool of child processes. Each child process might invoke a different
subroutine (that's up to you).

=head2 Methods

ActiveState::ProcessPool supports the following methods:

=over 4

=item new

   $pool = ActiveState::ProcessPool->new($size);

Create a new process pool of size $size. Does not actually fork() yet.

=item start

   $pool->start($slot, \&callback);

Creates a new child process in slot $slot. Each time you send a message to
slot $slot, the callback will be invoked in the child process.

Throws an exception if the slot already has a process.

The subroutine is passed an array reference containing key/value pairs passed
to it from the parent process (only simple scalars are supported).

   sub runner {
      my $kv = shift;
      # do something (possibly something time-consuming)
      return (a => 1, b => 2, c => 3);
   }

If the sub calls die() or croak(), the exception is trapped and returned to
the parent as if the return value had been:

   return ('$@' => $@);

This is on purpose.

=item start_long

   $pool->start_long($slot, \&longrun);

Creates a new child process in slot $slot, then immediately invokes
C<\&longrun> with no arguments. The process' STDIN and STDOUT will be
connected to the parent.

See L</Message Encoding> to understand the line protocol used.

=item send

   $pool->send($slot, @kv);

Encodes and sends a message to slot $slot, then returns immediately. It is
your responsibility to wait for the response from the $slot. (Hint: use
select() or poll().)

=item recv

   @kv = $pool->recv($slot);

Reads and decodes the message returned from $slot, blocking if a message is
not available. (Hint: use select() or poll().)

=item slots

   $n = $pool->slots;

Returns the number of running slots in the pool, or the actual slot indices in
list context.

=item size

   $n = $pool->size;

Returns the size of the process pool (the parameter passed to new()).

=item rfds

   @r = $pool->rfds;

Returns a list of file descriptors the parent can read from. There is one for
each running slot, representing the output from the child.

Note: returns the actual integer file descriptor, not a Perl filehandle.

=item wfds

   @w = $pool->wfds;

Returns a list of file descriptors the parent can write to. There is one for
each running slot, representing the input to the child.

Note: returns the actual integer file descriptor, not a Perl filehandle.

=item efds

   @e = $pool->efds;

Returns a list of file descriptors the parent can test for reading. If a file
descriptor in this list ever becomes readable, the child has died unexpectedly
and must be restarted.

=item rfd

=item wfd

=item efd

   $rfd = $o->rfd($slot);

Return the read, write, or error file descriptor for the specified slot.

=item write

=item read

   $pool->write($slot, $msg);
   $pool->read($slot, $buf, $len, $offset)

Similar to the Perl builtins, these read or write data to the underlying file
descriptor.

=item used

   $count = $pool->used($slot);

Returns the number of times the slot has been used. The counter increments
each time the parent C<write()>s to the child, and resets when the slot is
started.

=item pid

   $pid = $pool->pid($slot);

Returns the process id of the child process.

=back

=head2 Message Encoding

Messages are always passed as a set of key/value pairs. The encoding is very
simple: it begins with a 4-byte integer in native byte order specifying how
long the entire message is. The rest of the message is encoded to fit on a
single line, where each key/value pair is separated by a single space, and the
key and value are separated by an equals character.

For example, the example we have been using above would be encoded as follows:

   \0\0\0\014a=1 b=2 c=3

A more complicated example is how the following is encoded:

   $pool->write(1, a => '=', b => "\r\n", c => " ");

This is encoded as follows:

   \0\0\0\35a=?q?=3d b=?q?=0d=0a c=?q?=20

You can access the encoder and decoder by calling
ActiveState::ProcessPool::encode_kv() and
ActiveState::ProcessPool::decode_kv().
