package PureMessage::Policy::Virus;

use strict;

use base 'PureMessage::Policy::Plugin';

use PureMessage::Policy::Virus::Config '%config';
use PureMessage::Text::Wrap ();
use PureMessage::ListManager;
use PureMessage::Policy::Plugin::Utils qw(get_attachments
                                          set_attachments
                                          get_recipient
                                          get_matcher
                                          expand_localized_tmpl
                                          replace_nested_part
                                         );
use PureMessage::Policy::ArchiveInfo qw(archive_get_filenames);
use File::Basename;

use PureMessage::Sophos::SAVI::TFT qw(
        get_tft_mime_types
        get_tft_extensions
        );

use PureMessage::Sophos::SAVI::Utils qw(set_savi_cantscan_errors);

our $VERSION = '5.507';

# Decide which backends are installed, and suck in their methods.
my (@backends, $use_daemon);
for (grep $config{backend}{$_}{enabled}, keys %{$config{backend}}) {
    if ($config{backend}{$_}{daemon}) {
        require PureMessage::Policy::Virus::Client;
        $use_daemon++;
    }
    else {
        require PureMessage::_sys;
        my $b = __PACKAGE__ . "::$_";
        eval { PureMessage::_sys::load_module($b) };
        warn $@ if $@;
        push @backends, $@ ? () : $b;
    }
}

# Build a dispatch table for backends.
sub M_SCAN ()  { 0 }
sub M_CLEAN () { 1 }
my %backends;
for my $bk (@backends) {
    (my $bk_key = $bk) =~ s[^PureMessage::Policy::Virus::][];
    $backends{$bk} = $bk->impl($config{backend}{$bk_key});
}

# Build a hash of scan_failed_actions sorted by virus ID
my %scan_failed;
for (keys %{$config{scan_failed_action}}) {
    my $action = $config{scan_failed_action}{$_};
    foreach my $virus_id (@{$action->{id}}) {
        # silent support for old option-names
        $action->{action} = "deny" if $action->{action} eq "replace";

        $scan_failed{$virus_id}{action} = $action->{action};
        $scan_failed{$virus_id}{template} = $action->{template};
        $scan_failed{$virus_id}{desc} = $action->{description};
        $scan_failed{$virus_id}{id} = $virus_id;
    }
}

sub commands {
    my ($self, $cls, $logger) = @_;

    my @cmds = (
        pmx_suspect_attachment => \&pmx_suspect_attachment,
    );

    my @virus_cmds = (
        pmx_virus       => \&pmx_virus,
        pmx_virus_id    => \&pmx_virus_id,
        pmx_virus_clean => \&pmx_virus_clean,
        pmx_virus_cantscan => \&pmx_virus_cantscan,
    );

    if (@backends or $use_daemon) {
        push @cmds, @virus_cmds;
    }
    else {
         $logger->log('ERR', <<MSG) if grep /^pmx_virus/, keys %$cls;
No virus backends are loaded. Cannot implement 'pmx_virus' functionality.
MSG
    }

    return @cmds;
}

# $msg is a PureMessage::Policy::MessageObject thingy!
sub pmx_virus {
    my ($msg, $args) = @_;
    $msg->assert_event('eom');

    # Set a mark in the message_log to indicate we attempted 
    # to virus scan this message
    $msg->set_attr('mark_vs', undef);

    # Get all parts and flush to disk
    my (%parts, %map);
    $msg->enumerate_parts(sub {
        my $part = shift;
        my $filename = $part->body_filename;
        return 1 if not -s $filename;
        $parts{$filename} = $part;
        $map{$part->part_id}{filename} = $filename;
        return 1;
    });

    # Don't scan if there are no non-zero-length parts.
    my @files = keys %parts;
    return 0 unless @files;
    my %virus_ids;
    my @daemon_files = @files;

    # If we're using the daemon, tell it to begin scanning.
    my $conn = _daemon_scan_start($msg, @daemon_files) if $use_daemon;

    my $try = 0;
    while (1) {
        # Scan with in-process backends; if none, just skip to daemon mode
        last unless @backends;

        # First, scan each of the parts
        foreach my $filename (@files) {
            # Let each backend have a go
            my $part_id = $parts{$filename}->part_id if $parts{$filename};
            for my $bk (@backends) {
                my @bk_vlist = $backends{$bk}[M_SCAN]->($bk, $msg, $filename);

                # allow the ID if all cantscan.conf actions say so
                # (pmx_virus_clean takes care of whether to warn or deny)
                my @vlist = viruses_not_allowed($msg, @bk_vlist);
               
                # Update %virus_ids (the union of all viruses found)
                $virus_ids{$_}++ for @vlist;

                # Update %map (the map of which backends found viruses in
                # which parts) (except for the main part incl.headers)
                if ($parts{$filename} and @vlist) {
                    push @{$map{$part_id}{backend}}, $bk;
                    push @{$map{$part_id}{virus_id}}, @vlist;
                }
            }
        }

        # If we found nothing, try scanning the entire file (including headers).
        if (!$try and !(keys %virus_ids)) {
            $try++;
            @files = $msg->filename;
            next;
        }

        last;
    }

    # Discard any savi "corrupt" result from the final scan, unless we are
    # configured to be really strict and report that error
    if ($try and !$config{strict_corrupt}) {
        delete $virus_ids{SOPHOS_SAVI_FILE_CORRUPT};
    }
    
    $try = 0;
    while (1) {
        last unless $conn;

        # Read the response from the daemon (waiting if necessary)
        my %report = $conn->scan_file_report(@daemon_files);
        foreach my $filename (keys %report) {
            # Don't store into the map if this is the full file
            my $part_id = ($parts{$filename} ? $parts{$filename}->part_id : "main");

            if ($report{$filename}{problem}) {
                $msg->log("ERR", "[part $part_id] virus daemon failed: " .
                         $report{$filename}{descrip});
                # instead of using the 'onerror' setting, fire a new virus
                # error code:
                my $err = "SOPHOS_SAVI_FILE_TIMEOUT";
                $virus_ids{$err}++;
                push @{$map{$part_id}{virus_id}}, $err if $parts{$filename};
                push @{$map{$part_id}{backend}}, "daemon" if $parts{$filename}; # XXX
                next;
            }

            for my $bk (keys %{$report{$filename}}) {
                my $data = $report{$filename}{$bk};
                next unless ref $data eq 'ARRAY';

                # allow the ID if all all cantscan.conf actions say so
                # (pmx_virus_clean takes care of whether to warn or deny)
                my @vlist = viruses_not_allowed($msg, @$data);

                $virus_ids{$_}++ for @vlist;
                if ($parts{$filename} and @vlist) {
                    push @{$map{$parts{$filename}->part_id}{backend}}, $bk;
                    push @{$map{$parts{$filename}->part_id}{virus_id}}, @vlist;
                }
            }
        }

        $conn->disconnect;

        # If we found nothing, try scanning the entire file (including headers).
        if (!$try and !(keys %virus_ids)) {
            $try++;
            @daemon_files = $msg->filename;
            $conn = _daemon_scan_start($msg, @daemon_files);
            next;
        }

        last;
    }

    # Discard any savi "corrupt" result from the final scan, unless we are
    # configured to be really strict and report that error
    if ($try and !$config{strict_corrupt}) {
        delete $virus_ids{SOPHOS_SAVI_FILE_CORRUPT};
    }

    # Save the current list of viruses found (for internal use only)
    my @virus_ids = sort grep { $virus_ids{$_} } keys %virus_ids;
    $msg->set_attr('current_v', \@virus_ids);

    # Preserve the list of viruses from the last call, if any - note that the
    # %map is not also preserved - so it reflects the current message state
    if (my $old_ids = $msg->get_attr('mark_v')) {
        for my $virus_id (@$old_ids) {
            $virus_ids{$virus_id}++;
        }
    }

    @virus_ids = sort grep { $virus_ids{$_} } keys %virus_ids;
    $msg->set_attr('mark_v', \@virus_ids);
    $msg->set_attr('virus_map', \%map);
    $msg->template('VIRUS_IDS', join(', ', @virus_ids));
    $msg->minfo->append_items('Virus-Id' => $_) for @virus_ids;
    return @virus_ids ? 1 : 0;
}

sub pmx_virus_id {
    my ($msg, $args) = @_;
    my $ids = $msg->get_attr('mark_v');
    unless ($ids) {
        pmx_virus($msg, $args) or return 0;
        $ids = $msg->get_attr('mark_v');
    }
    return unless $ids and @$ids;
    my $matcher = get_matcher($msg, $args);
    return $matcher->(@$ids);
}

sub pmx_virus_clean {
    my ($msg, $args) = @_;
    my $repl = $args->param(0)->[0]; # the replacement file

    my (@bk_clean, @bk_cant, @vlist);
    for my $bk (@backends) {
        if ($backends{$bk}[M_CLEAN]) {
            push @bk_clean, $bk;
        }
        else {
            push @bk_cant, $bk;
        }
    }

    RETRY:
        my $map = $msg->get_attr('virus_map');
        unless ($map) {
            # Call pmx_virus if it hasn't been called already; exit if clean
            pmx_virus($msg, $args) or return 1;
            goto RETRY;
        };

        # Get a list of all infected parts and all viruses found
        my @part_ids = sort grep
            { defined $map->{$_}{virus_id} and @{$map->{$_}{virus_id}} } keys %$map;
        my $vlist = $msg->get_attr('mark_v');
        @vlist = @$vlist;

        # Try to clean each part in turn

    PART: 
        for my $part_id (@part_ids) {

            CLEAN: {

                my $part = $msg->part_nested($part_id, 1) or do {
                    $msg->log("DEBUG", "part id '$part_id' disappeared.");
                    $msg->log("DEBUG", "rescanning entire message");

                    # Re-scan the message, and return if it's clean:
                    pmx_virus($msg, $args) or return 1;

                    goto RETRY; # restart scanning at the first part
                };

                # If the ID is in a scan_failed section, 
                # do the action associated with the section
                # unless another virus was found in the part as well

                my $scan_fail_action;
                for my $virus_id (@{$map->{$part_id}{virus_id}}) {
                    if (my $action = $scan_failed{$virus_id}) {
                        if (!$scan_fail_action
                            # if there is more than one scan_failed ID,
                            # the replace action wins.
                            or ($scan_fail_action->{action} ne 'deny'
                                and $action->{action} eq 'deny')) {
                            $scan_fail_action = $action;
                            # XXX - potential problem here - 'deny' doesn't trump 'warn'
                        }
                    }
                    else {
                        $scan_fail_action = undef;
                        last;
                    }
                }
                if ($scan_fail_action) {
                    my $template = $scan_fail_action->{template};

                    my $f = $template !~ m#^[./]#
                        ? expand_localized_tmpl($msg, "virus.d/$template")
                        : $template;
                    if ($scan_fail_action->{action} ne 'deny') {
                        replace_nested_part($msg, 
                                            $part_id, 
                                            $f, 
                                            \&expand_templates, 
                                            warn_only=>1
                        );
                    }
                    else {
                        replace_nested_part($msg, 
                                            $part_id, 
                                            $f,
                                            \&expand_templates
                        );
                    }
                    next PART;
                }

                # If there's only one backend, and it can't clean, just
                # skip this whole section:
                last CLEAN if @backends == 1 and @bk_clean == 0 and not $use_daemon;

                # Clean with every backend that found a virus, and that
                # can clean viruses.
                for my $bk (@{$map->{$part_id}{backend}}) {
                    my $method = $backends{$bk}[M_CLEAN] or next;
                    my $res = $method->($bk, $msg, $map->{$part_id}{filename});
                    last CLEAN if $res eq 'deleted';
                }

                TRY: 
                    for my $try (0 .. 2) {
                        my $clean;

                        # Clean with all the capable backends. Try again if
                        # any of them doesn't think it's clean:
                        $clean = 0;
                        for my $bk (@bk_clean) {
                            my $result = $backends{$bk}[M_CLEAN]->(
                            $bk, $msg, $map->{$part_id}{filename});
                            last CLEAN if $result eq 'deleted';
                            $clean++ if $result eq 'novirus' or $result eq 'cleaned';
                        }
                        next TRY if $clean != @bk_clean;

                        # Now try cleaning with the daemon.
                        if ($use_daemon) {
                            my $filename = $map->{$part_id}{filename};
                            my $conn = PureMessage::Policy::Virus::Client->new;
                            $conn->connect;
                            $conn->clean_file_async($filename);
                            my %report = $conn->scan_file_report($filename);
                            $clean = 0;
                            my $backends = 0;
                            for my $bk (keys %{$report{$filename}}) {
                            my $res = $report{$filename}{$bk};
                            last CLEAN if $res eq 'deleted';
                            $clean++ if $res eq 'cleaned';
                            $backends++;
                            }
                            next TRY if $clean != $backends;
                        }

                        # Scan with the rest of the backends. Go to the next
                        # part if they don't find a virus.
                        $clean = 0;
                        for my $bk (@bk_cant) {
                            my @virus_ids = $backends{$bk}[M_SCAN]->(
                                        $bk, $msg, $map->{$part_id}{filename});
                            $clean++ unless @virus_ids;
                        }

                        if ($clean == @bk_cant) {
                            $part->set_modified;
                            $part->body_close;
                            next PART;
                        }

                        # There's no point trying again: all the cleaning
                        # scanners think it's clean, and the non-cleaning
                        # backends disagree. We have to delete the part.
                        last CLEAN;
                    }

                # If we got here, it means the cleaning backends can't
                # clean it after three tries, so we give up and delete the
                # part.
                last CLEAN;
            }

        # If we fall out here, we failed to clean the part, so
        # delete the part and go to the next one.
        my $f = $repl !~ m#^[./]#
            ? expand_localized_tmpl($msg, "virus.d/$repl")
            : $repl;

        replace_nested_part($msg, 
                            $part_id, 
                            $f,
                            \&expand_templates
        );
    }

    # Now we need to scan the message again to be sure there wasn't a virus in
    # the full message as well that we didn't notice before
    if (@part_ids) {
        pmx_virus($msg, $args) or return 1;
        $vlist = $msg->get_attr('current_v');
        @vlist = @$vlist;
    }

    # If no parts are listed in the map as having a virus, but we have a
    # non-null list of viruses found,
    if (!@part_ids and @vlist) {
        # ..then it's the entire message itself that's infected, so we'll
        # just nuke the entire thing.

        my $f = $repl !~ m#^[./]#
                ? expand_localized_tmpl($msg, "virus.d/$repl")
                : $repl;

        replace_nested_part($msg,
                            $msg->part_id,
                            $f,
                            \&expand_templates
        );

        return 1;
    }

    return 1; # cleaning always succeeds.
}


# Returns those viruses where the cantscan policy doesn't say to allow it
sub viruses_not_allowed {
    my $msg = shift;
    my @vlist;
    for my $virus_id (@_) {
        my $scan_fail_action = $scan_failed{$virus_id};
        if ($scan_fail_action) {
            $msg->log('NOTICE',
                "POLICY WARNING: Virus scanner was not able to scan an attachment" .
                " (id=$scan_fail_action->{id}, action=$scan_fail_action->{action})");
        }

        push @vlist, $virus_id unless
            $scan_fail_action and $scan_fail_action->{action} eq 'allow';
    }

    return @vlist;
}


sub expand_templates {
    my ($msg, $tmpl, $part_id) = @_;

    # Expand the templates
    use PureMessage::Template qw(expand_template);
    my $map = $msg->get_attr('virus_map');
    my @virus_ids;
    if (defined @{$map->{$part_id}{virus_id}}) {
        @virus_ids = @{$map->{$part_id}{virus_id}};
    }
    else {
        my $vlist = $msg->get_attr('current_v');
        @virus_ids = @$vlist;
    }
    my $desc = "";
    for (@virus_ids) {
        if (exists $scan_failed{$_} and exists $scan_failed{$_}{desc}) {
            $desc .= "$scan_failed{$_}{desc}\n";
        }
        else {
            $desc .= "No description associated with ID '$_'\n"
        }
    }
    chomp $desc;
    $tmpl = expand_template(
        $tmpl,
        vars => {
            header_subject	=> scalar $msg->header('subject'),
            virus_ids	    => join(', ', @virus_ids),
            qid		        => scalar $msg->minfo_value('Queue-Id'),
            desc            => $desc,
        },
    );
    return $tmpl;
}

sub _get_names {
    my $part = shift;
    my $names = [];

    my (undef, %ct) = $part->content_type;
    push @{$names}, $ct{name} if exists $ct{name};

    my $cd = $part->header_decoded('Content-Disposition');
    push @{$names}, $1 if $cd and $cd =~ /filename\s*=\s*\"?([^\r\n;\"]*)/i;

    return $names;
}

sub pmx_suspect_attachment {
    my ($msg,$args) = @_;
    my @matching;
    my @non_matching;

    # Clear the cantscan hash attribute
    my $cantscan_map = {'type' => 'tft', 'ids' => []};
    $msg->set_attr('cantscan_map', [$cantscan_map], 1);
        
    my $name_matcher = get_matcher($msg, $args, 'suspect-attachment-names');
    my $type_matcher = get_matcher($msg, $args, 'suspect-attachment-types');
    my $inspect_archives = $args->tag('inspect_archives');
    
    my @filtered = get_attachments($msg);
    $msg->enumerate_parts(sub {
        my $part = shift;
        return if @filtered && !(grep $part->part_id eq $_, @filtered);

        my @names = @{_get_names($part)};
        my @types = scalar $part->content_type;

        if($args->tag('tft')) {
            my $log_string = "pmx_suspect_attachment (part " . $part->part_id . "):";
            
            if (my $extensions = get_tft_extensions($part)) {
                push(@names, @$extensions);
                $msg->log("DEBUG", "$log_string extensions: " . join(', ', @$extensions));
            } else {
                $msg->log("DEBUG", "$log_string extensions: none found");
            }
            if (my $mimetypes = get_tft_mime_types($part)) {
                push(@names, @$mimetypes);
                $msg->log("DEBUG", "$log_string mime types: " . join(', ', @$mimetypes));
            } else {
                $msg->log("DEBUG", "$log_string mime types: none found");
            }

            # Store SAVI cantscan errors
            $msg->{per_message_savi_cantscan_errors} += set_savi_cantscan_errors(
                                                            $msg,
                                                            $part->part_id,
                                                            $part->{tft_info}->{sav_ids}->{$part->body_filename});

        # SAVI will look into archives anyway, so don't do it twice if TFT is enabled
        } elsif ($inspect_archives) {
            my $part_file = $part->body_filename;
            open my $fh, $part_file or die "could not open $part_file: $!";
            push @names, map (basename($_), archive_get_filenames($fh));
            close $fh;
        }

        if ($name_matcher->(@names) or $type_matcher->(@types)) {
            push @matching, $part->part_id;
        }
        else {
            push @non_matching, $part->part_id;
        }
    });
    
    return set_attachments($msg, \@matching, \@non_matching)
        || $msg->{per_message_savi_cantscan_errors};
}

sub pmx_virus_cantscan {
    my ($msg, $args) = @_;
    my $map = $msg->get_attr('virus_map');
    return 0 unless $map;
    
    my $scanfailed = 0;
    PART: foreach my $part_id (sort keys %$map) {
        for my $id (@{$map->{$part_id}{virus_id}}) {
            my $action;
            if ($action = $scan_failed{$id} 
                    and $action->{action} ne 'deny') {
                $scanfailed = 1;
            }
            else {
                $scanfailed = 0;
                last PART;
            }
        }
    }
    return $scanfailed;
}

#=============================================================================
# Support methods for virus backends.
#=============================================================================

# Handle some of the details with the daemon
sub _daemon_scan_start {
    return undef unless $use_daemon;
    my ($msg, @files) = @_;

    my $conn = PureMessage::Policy::Virus::Client->new;
    eval {
        $conn->connect;
        $conn->scan_file_async(@files);
    };
    if ($@) {
        $msg->log("ERR", "error connecting to virus daemon: $@");
        __PACKAGE__->v_error;
    };
    return $conn;
}

# Decide what to do about various error conditions
sub v_error {
    my $self = shift;
    my $etype = shift;

    # Error connecting to the virus daemon (only for those backends which do
    # connect to daemons, obviously)
    if ($etype eq 'connect') {
    }
    # Virus scanner is dead, or died. For command-line scanners, the command
    # failed to fork/exec.
    elsif ($etype eq 'dead') {
    }
    # Virus daemon returned 'busy' error (i.e., too many queued connections).
    elsif ($etype eq 'busy') {
    }
    # Message took too long to scan (in some way)
    elsif ($etype eq 'timeout') {
	my $timeout = shift;

	# Too many wall-clock seconds
	if ($timeout eq 'scan_limit') {
	}
	# Virus scanner appeared inactive
	elsif ($etype eq 'inactive_limit') {
	}
	# for future expansion:
	# elsif ($etype eq 'cpu_user_limit') { }
	# elsif ($etype eq 'cpu_sys_limit') { }
    }
    
    if ($config{onerror} eq 'tempfail') {
        die "engine failure - tempfailing message\n";
    }
}

1;
