package PureMessage::Policy::Virus::Config;

use strict;
use PureMessage::Config qw(expand_filename);
use base qw(PureMessage::Config::Validator Exporter);

our @EXPORT_OK = '%config';
our %config;

my @typedefs = (
    ['#scan_poll_interval',     \&v_scan_poll_interval, undef],
);


my %nodes = (
    # OPTION_NAME                       => [ TYPE, DEFAULT ],

    onerror                             => ['#enum passthru tempfail', 'tempfail'],

    # "temporary" workaround for libsavi being too picky about message corruption
    strict_corrupt                      => ['#bool', 0],

    'daemon.port'                       => ['#port', 'local:/tmp/pmx-vscan.sock'],
    'daemon.concurrency_limit'          => ['#int', \&d_cclim],
    'daemon.reinit_limit'               => ['#int', 1024],
    'daemon.queue_limit'                => ['#int', 4096],
    'daemon.inactive_limit'             => ['#duration', '2m'],
    'daemon.scan_limit'                 => ['#duration', '5m'],
    'daemon.respawn_wait'               => ['#duration', '5m'],

    # intentionally undocumented and not included in conf file
    'daemon.scan_poll_interval'         => ['#duration #scan_poll_interval', '15s'],

    'scan_failed_action#*.id'           => ['#list #str'],
    'scan_failed_action#*.action'       => ['#enum allow deny warn replace', 'warn'],
    'scan_failed_action#*.template'     => ['#str', 'cantscan.tmpl'],
    'scan_failed_action#*.description'  => ['#str'],
);

refresh();

sub new {
    my $self = shift;
    require PureMessage::Config::Validator;
    my $p = ActiveState::Config->new(strict => 1,
				     allow_includes => 1);
    define_core_validators($self, $p);

    PureMessage::Policy::Virus::Config->validate($p);
    $p->parse_file(expand_filename("etc/virus.conf"), @_);

    $p;
}

sub define_core_validators {
    my ($pkg, $p) = @_;
    PureMessage::Config::Validator->define_core_validators($p);
    for my $def (@typedefs) {
	my ($type, $validator, $formatter) = @$def;
	$p->typedecl($type, $validator, $formatter);
    }
}

sub refresh {
    %config = new(undef)->as_hash;
}

sub d_cclim {
    require PureMessage::_sys;
    return PureMessage::_sys::available_cpus() * 2;
}

sub validate {
    my ($pkg, $p) = @_;

    for my $key (keys %nodes) {
        my ($type, $default, $sub) = @{$nodes{$key}};
        $p->typedecl($key, $type)   if defined $type;
        $p->default($key, $default) if defined $default;
    }

    $pkg->include_sub_validators($p);
}

# Local types

sub v_scan_poll_interval {
    my ($p, $path, $node, $value, $append, $ctx, $ts) = @_;

    # NOTE: the select() timeout should not be too high, otherwise if a
    # child goes into an infinite loop, it could go unnoticed if no other
    # handles are readable or writable; likewise sub-second (or constant)
    # polling will make the daemon go crazy
    my ($min, $max) = (1, 15);

    if ($value < $min) {
        $_[&ActiveState::Config::Types::VALUE_IX] = $min;
        $_[&ActiveState::Config::Types::CTX_IX] = 1;
    }
    elsif ($value > $max) {
        $_[&ActiveState::Config::Types::VALUE_IX] = $max;
        $_[&ActiveState::Config::Types::CTX_IX] = 1;
    }

    1;
}


1;
